@@ -69,15 +69,12 @@ impl<T: EventListener> Context<T> {
         let has_updated = self.renderable_content.selection_range != selection_range;
 
         if has_updated {
-            let mut terminal = self.terminal.lock();
-            let display_offset = terminal.display_offset();
-            terminal.update_selection_damage(selection_range, display_offset);
-            drop(terminal);
-
-            // Mark pending update as dirty so the selection change is rendered
+            // Selection is a UI-level change, so we use UI damage tracking
+            // Use full damage for selections since they can span multiple lines
+            // and change frequently (during dragging)
             self.renderable_content
                 .pending_update
-                .invalidate_full(&self.terminal);
+                .set_ui_damage(rio_backend::event::TerminalDamage::Full);
         }
 
         self.renderable_content.selection_range = selection_range;
@@ -85,10 +82,16 @@ impl<T: EventListener> Context<T> {
 
     #[inline]
     pub fn set_hyperlink_range(&mut self, hyperlink_range: Option<SelectionRange>) {
+        let old_hyperlink = self.renderable_content.hyperlink_range;
+
+        if old_hyperlink != hyperlink_range {
+            // For hyperlinks, use full damage as they're less frequent
+            self.renderable_content
+                .pending_update
+                .set_ui_damage(rio_backend::event::TerminalDamage::Full);
+        }
+
         self.renderable_content.hyperlink_range = hyperlink_range;
-        self.renderable_content
-            .pending_update
-            .invalidate_full(&self.terminal);
     }
 
     #[inline]
@@ -283,8 +286,13 @@ impl<T: EventListener + Clone + std::marker::Send + 'static> ContextManager<T> {
             }
         }
 
-        let machine =
-            Machine::new(Arc::clone(&terminal), pty, event_proxy.clone(), window_id)?;
+        let machine = Machine::new(
+            Arc::clone(&terminal),
+            pty,
+            event_proxy.clone(),
+            window_id,
+            route_id,
+        )?;
         let channel = machine.channel();
         let io_thread = if config.spawn_performer {
             Some(machine.spawn())
