@@ -16,7 +16,6 @@ use crate::context::ContextManager;
 use crate::crosswords::grid::row::Row;
 use crate::crosswords::pos::{Column, Line, Pos};
 use crate::crosswords::square::{Flags, Square};
-use crate::screen::hint::HintMatches;
 use navigation::ScreenNavigation;
 use rio_backend::config::colors::term::TermColors;
 use rio_backend::config::colors::{
@@ -42,13 +41,14 @@ pub struct Search {
 
 pub struct Renderer {
     is_vi_mode_enabled: bool,
+    is_game_mode_enabled: bool,
     draw_bold_text_with_light_colors: bool,
     use_drawable_chars: bool,
     pub named_colors: Colors,
     pub colors: List,
     pub navigation: ScreenNavigation,
     unfocused_split_opacity: f32,
-    last_active: Option<slotmap::DefaultKey>,
+    last_active: Option<usize>,
     pub config_has_blinking_enabled: bool,
     pub config_blinking_interval: u64,
     ignore_selection_fg_color: bool,
@@ -116,6 +116,7 @@ impl Renderer {
             font_cache: FontCache::new(),
             font_context: font_context.clone(),
             char_cache: CharCache::new(),
+            is_game_mode_enabled: config.renderer.strategy.is_game(),
         };
 
         // Pre-populate font cache with common characters for better performance
@@ -235,6 +236,15 @@ impl Renderer {
     }
 
     #[inline]
+    #[allow(clippy::too_many_arguments)]
+    /// Check if a position is within any hint match
+    fn is_position_in_hint_matches(
+        matches: &[rio_backend::crosswords::search::Match],
+        pos: Pos,
+    ) -> bool {
+        matches.iter().any(|m| m.contains(&pos))
+    }
+
     #[allow(clippy::too_many_arguments)]
     fn create_line(
         &mut self,
@@ -244,14 +254,13 @@ impl Renderer {
         line_opt: Option<usize>,
         line: Line,
         renderable_content: &RenderableContent,
-        search_hints: &mut Option<HintMatches>,
+        hint_matches: Option<&[rio_backend::crosswords::search::Match]>,
         focused_match: &Option<RangeInclusive<Pos>>,
         term_colors: &TermColors,
         is_active: bool,
     ) {
         // let start = std::time::Instant::now();
         let cursor = &renderable_content.cursor;
-        let hyperlink_range = renderable_content.hyperlink_range;
         let selection_range = renderable_content.selection_range;
         let columns: usize = row.len();
         let mut content = String::with_capacity(columns);
@@ -270,25 +279,33 @@ impl Renderer {
                 continue;
             }
 
-            let (mut style, square_content) =
+            let (mut style, mut square_content) =
                 if has_cursor && column == cursor.state.pos.col {
                     self.create_cursor_style(square, cursor, is_active, term_colors)
                 } else {
                     self.create_style(square, term_colors)
                 };
 
-            if hyperlink_range.is_some()
-                && square.hyperlink().is_some()
-                && hyperlink_range
-                    .unwrap()
-                    .contains(Pos::new(line, Column(column)))
-            {
+            // Apply underline for hyperlinks (OSC 8) or highlighted hints (hover)
+            let should_underline = square.hyperlink().is_some() || {
+                if let Some(highlighted_hint) = &renderable_content.highlighted_hint {
+                    let current_pos = Pos::new(line, Column(column));
+                    highlighted_hint.start <= current_pos
+                        && current_pos <= highlighted_hint.end
+                } else {
+                    false
+                }
+            };
+
+            if should_underline {
                 style.decoration =
                     Some(FragmentStyleDecoration::Underline(UnderlineInfo {
                         is_doubled: false,
                         shape: UnderlineShape::Regular,
                     }));
-            } else if selection_range.is_some()
+            }
+
+            if selection_range.is_some()
                 && selection_range
                     .unwrap()
                     .contains(Pos::new(line, Column(column)))
@@ -299,10 +316,11 @@ impl Renderer {
                     self.named_colors.selection_foreground
                 };
                 style.background_color = Some(self.named_colors.selection_background);
-            } else if search_hints.is_some()
-                && search_hints
-                    .as_mut()
-                    .is_some_and(|search| search.advance(Pos::new(line, Column(column))))
+            } else if hint_matches.is_some()
+                && Self::is_position_in_hint_matches(
+                    hint_matches.unwrap(),
+                    Pos::new(line, Column(column)),
+                )
             {
                 let is_focused = focused_match
                     .as_ref()
@@ -318,6 +336,42 @@ impl Renderer {
                 }
             }
 
+            // Check for hint labels at this position
+            if let Some(hint_label) = self.find_hint_label_at_position(
+                renderable_content,
+                Pos::new(line, Column(column)),
+            ) {
+                // Override character with hint label character if available
+                if let Some(label_char) = hint_label.label.first() {
+                    square_content = *label_char;
+                }
+
+                // Apply hint label styling
+                if hint_label.is_first {
+                    // Use configurable hint colors
+                    style.color = self.named_colors.hint_foreground;
+                    style.background_color = Some(self.named_colors.hint_background);
+                } else {
+                    // End colors: use same foreground, slightly dimmed background
+                    style.color = self.named_colors.hint_foreground;
+                    let mut dimmed_bg = self.named_colors.hint_background;
+                    // Dim the background slightly for continuation labels
+                    dimmed_bg[0] *= 0.8;
+                    dimmed_bg[1] *= 0.8;
+                    dimmed_bg[2] *= 0.8;
+                    style.background_color = Some(dimmed_bg);
+                }
+
+                // Make hint labels bold for better visibility
+                use rio_backend::sugarloaf::font_introspector::{Attributes, Weight};
+                let current_attrs = style.font_attrs;
+                style.font_attrs = Attributes::new(
+                    current_attrs.stretch(),
+                    Weight::BOLD,
+                    current_attrs.style(),
+                );
+            }
+
             if !is_active {
                 style.color[3] = self.unfocused_split_opacity;
                 if let Some(mut background_color) = style.background_color {
@@ -750,7 +804,6 @@ impl Renderer {
         &mut self,
         sugarloaf: &mut Sugarloaf,
         context_manager: &mut ContextManager<EventProxy>,
-        hints: &mut Option<HintMatches>,
         focused_match: &Option<RangeInclusive<Pos>>,
     ) {
         // let start = std::time::Instant::now();
@@ -772,7 +825,7 @@ impl Renderer {
         }
 
         for (key, grid_context) in grid.contexts_mut().iter_mut() {
-            let is_active = active_key == key;
+            let is_active = &active_key == key;
             let context = grid_context.context_mut();
 
             let mut has_ime = false;
@@ -790,7 +843,7 @@ impl Renderer {
                     context.renderable_content.cursor.content_ref;
             }
 
-            let force_full_damage = has_active_changed || is_active && hints.is_some();
+            let force_full_damage = has_active_changed || self.is_game_mode_enabled;
 
             // Check if we need to render
             if !context.renderable_content.pending_update.is_dirty() && !force_full_damage
@@ -799,35 +852,62 @@ impl Renderer {
                 continue;
             }
 
-            // Take the pending snapshot
-            let terminal_snapshot =
-                match context.renderable_content.pending_update.take_snapshot() {
-                    Some(snapshot) => snapshot,
-                    None if force_full_damage => {
-                        // Force full damage case - create a fresh snapshot
-                        let mut terminal = context.terminal.lock();
-                        let snapshot = TerminalSnapshot {
-                            colors: terminal.colors,
-                            display_offset: terminal.display_offset(),
-                            blinking_cursor: terminal.blinking_cursor,
-                            visible_rows: terminal.visible_rows(),
-                            cursor: terminal.cursor(),
-                            damage: TerminalDamage::Full,
-                            columns: terminal.columns(),
-                            screen_lines: terminal.screen_lines(),
-                        };
-                        terminal.reset_damage();
-                        drop(terminal);
-                        snapshot
-                    }
-                    None => {
-                        // No pending update and not forcing
-                        continue;
-                    }
+            // Get UI damage before resetting
+            let ui_damage = context.renderable_content.pending_update.take_ui_damage();
+            context.renderable_content.pending_update.reset();
+
+            // Compute snapshot at render time
+            let terminal_snapshot = {
+                let mut terminal = context.terminal.lock();
+
+                // Get damage from terminal
+                let terminal_damage = if force_full_damage {
+                    Some(TerminalDamage::Full)
+                } else {
+                    terminal.peek_damage_event()
                 };
 
-            // Process the snapshot
-            // let _duration = start.elapsed();
+                // Merge terminal damage with UI damage
+                let damage = match (terminal_damage, ui_damage) {
+                    (Some(TerminalDamage::Full), _) | (_, Some(TerminalDamage::Full)) => {
+                        TerminalDamage::Full
+                    }
+                    (Some(term), Some(ui)) => {
+                        // Merge partial damages
+                        match (term, ui) {
+                            (
+                                TerminalDamage::Partial(mut lines1),
+                                TerminalDamage::Partial(lines2),
+                            ) => {
+                                lines1.extend(lines2);
+                                TerminalDamage::Partial(lines1)
+                            }
+                            _ => TerminalDamage::Full,
+                        }
+                    }
+                    (Some(damage), None) => damage,
+                    (None, Some(damage)) => damage,
+                    (None, None) => TerminalDamage::Full,
+                };
+
+                let snapshot = TerminalSnapshot {
+                    colors: terminal.colors,
+                    display_offset: terminal.display_offset(),
+                    blinking_cursor: terminal.blinking_cursor,
+                    visible_rows: terminal.visible_rows(),
+                    cursor: terminal.cursor(),
+                    damage,
+                    columns: terminal.columns(),
+                    screen_lines: terminal.screen_lines(),
+                };
+                terminal.reset_damage();
+                drop(terminal);
+
+                snapshot
+            };
+
+            // Get hint matches from renderable content
+            let hint_matches = context.renderable_content.hint_matches.as_deref();
 
             // Update cursor state from snapshot
             context.renderable_content.cursor.state = terminal_snapshot.cursor;
@@ -945,7 +1025,7 @@ impl Renderer {
                             None,
                             Line((i as i32) - terminal_snapshot.display_offset as i32),
                             &context.renderable_content,
-                            hints,
+                            hint_matches,
                             focused_match,
                             &terminal_snapshot.colors,
                             is_active,
@@ -974,7 +1054,7 @@ impl Renderer {
                                         - terminal_snapshot.display_offset as i32,
                                 ),
                                 &context.renderable_content,
-                                hints,
+                                hint_matches,
                                 focused_match,
                                 &terminal_snapshot.colors,
                                 is_active,
@@ -1024,4 +1104,160 @@ impl Renderer {
 
         // let _duration = start.elapsed();
     }
+
+    /// Find hint label at the specified position
+    fn find_hint_label_at_position<'a>(
+        &self,
+        renderable_content: &'a RenderableContent,
+        pos: Pos,
+    ) -> Option<&'a crate::context::renderable::HintLabel> {
+        renderable_content
+            .hint_labels
+            .iter()
+            .find(|label| label.position == pos)
+    }
+}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use rio_backend::crosswords::pos::{Column, Line, Pos};
+
+    #[test]
+    fn test_is_position_in_hint_matches() {
+        let matches = vec![
+            Pos::new(Line(0), Column(0))..=Pos::new(Line(0), Column(4)),
+            Pos::new(Line(1), Column(5))..=Pos::new(Line(1), Column(9)),
+            Pos::new(Line(5), Column(10))..=Pos::new(Line(5), Column(15)),
+        ];
+
+        // Test positions inside matches
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(0), Column(0))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(0), Column(2))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(0), Column(4))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(1), Column(5))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(1), Column(7))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(1), Column(9))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(5), Column(10))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(5), Column(15))
+        ));
+
+        // Test positions outside matches
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(0), Column(5))
+        ));
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(1), Column(4))
+        ));
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(1), Column(10))
+        ));
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(2), Column(0))
+        ));
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(5), Column(9))
+        ));
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(5), Column(16))
+        ));
+    }
+
+    #[test]
+    fn test_empty_hint_matches() {
+        let matches: Vec<rio_backend::crosswords::search::Match> = vec![];
+
+        // Any position should return false for empty matches
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(0), Column(0))
+        ));
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(10), Column(20))
+        ));
+    }
+
+    #[test]
+    fn test_single_character_match() {
+        let matches = vec![Pos::new(Line(3), Column(7))..=Pos::new(Line(3), Column(7))];
+
+        // Test the exact position
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(3), Column(7))
+        ));
+
+        // Test adjacent positions
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(3), Column(6))
+        ));
+        assert!(!Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(3), Column(8))
+        ));
+    }
+
+    #[test]
+    fn test_overlapping_matches() {
+        // In practice, matches shouldn't overlap, but let's test the behavior
+        let matches = vec![
+            Pos::new(Line(2), Column(5))..=Pos::new(Line(2), Column(10)),
+            Pos::new(Line(2), Column(8))..=Pos::new(Line(2), Column(12)),
+        ];
+
+        // Test positions in the overlap
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(2), Column(8))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(2), Column(9))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(2), Column(10))
+        ));
+
+        // Test positions in non-overlapping parts
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(2), Column(5))
+        ));
+        assert!(Renderer::is_position_in_hint_matches(
+            &matches,
+            Pos::new(Line(2), Column(12))
+        ));
+    }
 }
