@@ -13,11 +13,11 @@ use crate::monitor::MonitorHandle;
 use crate::window::{BadIcon, Icon, Window, WindowAttributes};
 
 /// Window Handle type used by Win32 API
-pub type HWND = isize;
+pub type HWND = *mut c_void;
 /// Menu Handle type used by Win32 API
-pub type HMENU = isize;
+pub type HMENU = *mut c_void;
 /// Monitor Handle type used by Win32 API
-pub type HMONITOR = isize;
+pub type HMONITOR = *mut c_void;
 
 /// Describes a system-drawn backdrop material of a window.
 ///
@@ -304,6 +304,13 @@ pub trait WindowExtWindows {
     /// Supported starting with Windows 11 Build 22000.
     fn set_corner_preference(&self, preference: CornerPreference);
 
+    /// Cloaks the window such that it is not visible to the user. The window is still composed by DWM.
+    ///
+    /// A cloaked window will not appear on Taskbar. Cloaking is particularly useful when you want a
+    /// window to complete its UI layout, sizing, etc. without flickering in front of the user.
+    /// Not supported on Windows 7 and earlier.
+    fn set_cloaked(&self, cloaked: bool);
+
     /// Get the raw window handle for this [`Window`] without checking for thread affinity.
     ///
     /// Window handles in Win32 have a property called "thread affinity" that ties them to their
@@ -408,6 +415,11 @@ impl WindowExtWindows for Window {
         self.window.set_corner_preference(preference)
     }
 
+    #[inline]
+    fn set_cloaked(&self, cloaked: bool) {
+        self.window.set_cloaked(cloaked)
+    }
+
     unsafe fn window_handle_any_thread(
         &self,
     ) -> Result<raw_window_handle::WindowHandle<'_>, raw_window_handle::HandleError> {
@@ -536,6 +548,11 @@ pub trait WindowAttributesExtWindows {
     ///
     /// Supported starting with Windows 11 Build 22000.
     fn with_corner_preference(self, corners: CornerPreference) -> Self;
+
+    /// Cloaks the window such that it is not visible to the user. The window is still composed by DWM.
+    ///
+    /// Not supported on Windows 7 and earlier.
+    fn with_cloaked(self, cloaked: bool) -> Self;
 }
 
 impl WindowAttributesExtWindows for WindowAttributes {
@@ -623,6 +640,12 @@ impl WindowAttributesExtWindows for WindowAttributes {
         self.platform_specific.corner_preference = Some(corners);
         self
     }
+
+    #[inline]
+    fn with_cloaked(mut self, cloaked: bool) -> Self {
+        self.platform_specific.cloaked = cloaked;
+        self
+    }
 }
 
 /// Additional methods on `MonitorHandle` that are specific to Windows.
