@@ -1,8 +1,17 @@
 use crate::context::Context;
+use tracing::debug;
 
 use super::atlas::*;
+use super::ContentType;
 use super::*;
 
+#[derive(Clone, Copy, Debug, PartialEq, Default)]
+pub enum AtlasKind {
+    #[default]
+    Mask, // R8 format for alpha masks
+    Color, // RGBA format for color glyphs
+}
+
 #[derive(Default)]
 pub struct Entry {
     allocated: bool,
@@ -14,6 +23,8 @@ pub struct Entry {
     width: u16,
     /// Height of the image.
     height: u16,
+    /// Which atlas this entry belongs to
+    atlas_kind: AtlasKind,
 }
 
 pub struct Atlas {
@@ -21,14 +32,35 @@ pub struct Atlas {
     buffer: Vec<u8>,
     fresh: bool,
     dirty: bool,
+    channels: usize, // 1 for mask, 4 for color
+}
+
+impl Atlas {
+    fn new(kind: AtlasKind) -> Self {
+        let channels = match kind {
+            AtlasKind::Mask => 1,
+            AtlasKind::Color => 4, // Always 4 for Rgba8Unorm
+        };
+
+        Self {
+            alloc: AtlasAllocator::new(SIZE, SIZE),
+            buffer: vec![0; SIZE as usize * SIZE as usize * channels],
+            fresh: true,
+            dirty: false,
+            channels,
+        }
+    }
 }
 
 pub struct ImageCache {
     pub entries: Vec<Entry>,
-    atlas: Atlas,
+    mask_atlas: Atlas,
+    color_atlas: Atlas,
     max_texture_size: u16,
-    texture: wgpu::Texture,
-    pub texture_view: wgpu::TextureView,
+    mask_texture: wgpu::Texture,
+    color_texture: wgpu::Texture,
+    pub mask_texture_view: wgpu::TextureView,
+    pub color_texture_view: wgpu::TextureView,
 }
 
 #[inline]
@@ -41,14 +73,14 @@ pub fn buffer_size(width: u32, height: u32) -> Option<usize> {
 pub const SIZE: u16 = 2048;
 
 impl ImageCache {
-    /// Creates a new image cache.
+    /// Creates a new image cache with dual atlases.
     pub fn new(context: &Context) -> Self {
         let device = &context.device;
-        // let max_texture_size = max_texture_size.clamp(1024, 8192);
         let max_texture_size = SIZE;
 
-        let texture = device.create_texture(&wgpu::TextureDescriptor {
-            label: Some("rich_text create texture"),
+        // Create mask texture (R8 format for alpha masks)
+        let mask_texture = device.create_texture(&wgpu::TextureDescriptor {
+            label: Some("rich_text mask atlas"),
             size: wgpu::Extent3d {
                 width: SIZE as u32,
                 height: SIZE as u32,
@@ -56,29 +88,42 @@ impl ImageCache {
             },
             view_formats: &[],
             dimension: wgpu::TextureDimension::D2,
-            format: wgpu::TextureFormat::Rgba8Unorm,
+            format: wgpu::TextureFormat::R8Unorm,
             usage: wgpu::TextureUsages::COPY_DST | wgpu::TextureUsages::TEXTURE_BINDING,
             mip_level_count: 1,
             sample_count: 1,
         });
-        let texture_view = texture.create_view(&wgpu::TextureViewDescriptor::default());
+        let mask_texture_view =
+            mask_texture.create_view(&wgpu::TextureViewDescriptor::default());
 
-        let alloc = AtlasAllocator::new(max_texture_size, max_texture_size);
+        // Create color texture (RGBA8 format for color glyphs - simpler than f16)
+        let color_texture_format = wgpu::TextureFormat::Rgba8Unorm;
+        let color_texture = device.create_texture(&wgpu::TextureDescriptor {
+            label: Some("rich_text color atlas"),
+            size: wgpu::Extent3d {
+                width: SIZE as u32,
+                height: SIZE as u32,
+                depth_or_array_layers: 1,
+            },
+            view_formats: &[],
+            dimension: wgpu::TextureDimension::D2,
+            format: color_texture_format,
+            usage: wgpu::TextureUsages::COPY_DST | wgpu::TextureUsages::TEXTURE_BINDING,
+            mip_level_count: 1,
+            sample_count: 1,
+        });
+        let color_texture_view =
+            color_texture.create_view(&wgpu::TextureViewDescriptor::default());
 
         Self {
             entries: Vec::new(),
-            atlas: Atlas {
-                alloc,
-                buffer: vec![
-                    0u8;
-                    max_texture_size as usize * max_texture_size as usize * 4
-                ],
-                fresh: true,
-                dirty: true,
-            },
+            mask_atlas: Atlas::new(AtlasKind::Mask),
+            color_atlas: Atlas::new(AtlasKind::Color), // Always 4 bytes per pixel for Rgba8Unorm
             max_texture_size,
-            texture_view,
-            texture,
+            mask_texture,
+            color_texture,
+            mask_texture_view,
+            color_texture_view,
         }
     }
 
@@ -87,6 +132,11 @@ impl ImageCache {
         let width = request.width;
         let height = request.height;
 
+        // Reject zero-sized images
+        if width == 0 || height == 0 {
+            return None;
+        }
+
         // Check buffer size
         buffer_size(width as u32, height as u32)?;
 
@@ -95,27 +145,28 @@ impl ImageCache {
             return None;
         }
 
-        let atlas_data = self.atlas.alloc.allocate(width, height);
-        // if atlas_data.is_none() {
-        // return None;
-        // Grow atlas to fit
-        // self.max_texture_size += SIZE;
-        // self.atlas.fresh = true;
-        // self.atlas.dirty = true;
-        // self.entries.clear();
-        // println!("{:?}", self.max_texture_size);
-        // self.atlas.alloc = AtlasAllocator::new(self.max_texture_size, self.max_texture_size);
-        // self.atlas.buffer = vec![
-        //     0u8;
-        //     self.max_texture_size as usize * self.max_texture_size as usize * 4
-        // ];
-        // atlas_data = self.atlas.alloc.allocate(width, height);
-        // println!("{:?}", atlas_data);
+        // Choose the appropriate atlas based on content type
+        let atlas_kind = match request.content_type {
+            ContentType::Mask => AtlasKind::Mask,
+            ContentType::Color => AtlasKind::Color,
+        };
+
+        let atlas = match atlas_kind {
+            AtlasKind::Mask => &mut self.mask_atlas,
+            AtlasKind::Color => &mut self.color_atlas,
+        };
+
+        let atlas_data = atlas.alloc.allocate(width, height);
+
+        // Log cache miss when allocation fails
+        if atlas_data.is_none() {
+            debug!(
+                "ImageCache allocation failed for {}x{} - {:?} atlas full",
+                width, height, atlas_kind
+            );
+            return None;
+        }
 
-        // if self.max_texture_size > MAX_SIZE {
-        //     println!("should try to grow or reset atlas");
-        // }
-        // }
         let (x, y) = atlas_data?;
         let entry_index = self.entries.len();
         self.entries.push(Entry {
@@ -124,19 +175,25 @@ impl ImageCache {
             y,
             width,
             height,
+            atlas_kind,
         });
+
         if let Some(data) = request.data() {
             fill(
-                x,
-                y,
-                width,
-                height,
+                FillParams {
+                    x,
+                    y,
+                    width,
+                    _height: height,
+                    target_width: self.max_texture_size,
+                    channels: atlas.channels,
+                },
                 data,
-                self.max_texture_size,
-                &mut self.atlas.buffer,
+                &mut atlas.buffer,
             );
-            self.atlas.dirty = true;
+            atlas.dirty = true;
         }
+
         ImageId::new(entry_index as u32, request.has_alpha)
     }
 
@@ -151,13 +208,23 @@ impl ImageCache {
             return None;
         }
 
-        self.atlas.alloc.deallocate(entry.x, entry.y, entry.width);
+        let atlas = match entry.atlas_kind {
+            AtlasKind::Mask => &mut self.mask_atlas,
+            AtlasKind::Color => &mut self.color_atlas,
+        };
+
+        atlas.alloc.deallocate(entry.x, entry.y, entry.width);
         entry.allocated = false;
         Some(())
     }
 
     /// Retrieves the image for the specified handle and updates the epoch.
     pub fn get(&self, handle: &ImageId) -> Option<ImageLocation> {
+        // Empty images have no location (for zero-sized glyphs)
+        if handle.is_empty() {
+            return None;
+        }
+
         let entry = self.entries.get(handle.index())?;
         if !entry.allocated {
             return None;
@@ -172,8 +239,25 @@ impl ImageCache {
         })
     }
 
+    /// Clears all entries and resets the atlas. Used when fonts change.
+    pub fn clear_atlas(&mut self) {
+        // Clear all entries
+        self.entries.clear();
+
+        // Reset both atlases
+        self.mask_atlas = Atlas::new(AtlasKind::Mask);
+        self.color_atlas = Atlas::new(AtlasKind::Color);
+
+        tracing::info!("Dual atlases cleared due to font change");
+    }
+
     /// Returns true if the image is valid.
     pub fn is_valid(&self, image: ImageId) -> bool {
+        // Empty images are always valid (for zero-sized glyphs)
+        if image.is_empty() {
+            return true;
+        }
+
         if let Some(entry) = self.entries.get(image.index()) {
             entry.allocated
         } else {
@@ -201,97 +285,84 @@ impl ImageCache {
     //         atlas.dirty = true;
     //     Some(())
     // }
-
     #[inline]
     pub fn process_atlases(&mut self, context: &mut Context) {
-        if !self.atlas.dirty {
-            return;
-        }
-        if self.atlas.fresh {
+        // Process mask atlas
+        if self.mask_atlas.dirty {
             let texture_size = wgpu::Extent3d {
-                width: (self.max_texture_size).into(),
-                height: (self.max_texture_size).into(),
+                width: self.max_texture_size as u32,
+                height: self.max_texture_size as u32,
                 depth_or_array_layers: 1,
             };
-            let new_texture = context.device.create_texture(&wgpu::TextureDescriptor {
-                size: texture_size,
-                mip_level_count: 1,
-                sample_count: 1,
-                dimension: wgpu::TextureDimension::D2,
-                format: wgpu::TextureFormat::Rgba8Unorm,
-                usage: wgpu::TextureUsages::TEXTURE_BINDING
-                    | wgpu::TextureUsages::COPY_DST,
-                label: Some("rich_text::fresh atlas"),
-                view_formats: &[],
-            });
 
             context.queue.write_texture(
-                // Tells wgpu where to copy the pixel data
-                wgpu::ImageCopyTexture {
-                    texture: &new_texture,
+                wgpu::TexelCopyTextureInfo {
+                    texture: &self.mask_texture,
                     mip_level: 0,
                     origin: wgpu::Origin3d::ZERO,
                     aspect: wgpu::TextureAspect::All,
                 },
-                // The actual pixel data
-                &self.atlas.buffer,
-                // The layout of the texture
-                wgpu::ImageDataLayout {
+                &self.mask_atlas.buffer,
+                wgpu::TexelCopyBufferLayout {
                     offset: 0,
-                    bytes_per_row: Some((self.max_texture_size * 4).into()),
-                    rows_per_image: Some((self.max_texture_size).into()),
+                    bytes_per_row: Some(
+                        self.max_texture_size as u32 * self.mask_atlas.channels as u32,
+                    ),
+                    rows_per_image: Some(self.max_texture_size as u32),
                 },
                 texture_size,
             );
 
-            self.texture = new_texture;
-            self.texture_view = self
-                .texture
-                .create_view(&wgpu::TextureViewDescriptor::default());
-        } else {
+            self.mask_atlas.fresh = false;
+            self.mask_atlas.dirty = false;
+        }
+
+        // Process color atlas
+        if self.color_atlas.dirty {
             let texture_size = wgpu::Extent3d {
-                width: (self.max_texture_size).into(),
-                height: (self.max_texture_size).into(),
+                width: self.max_texture_size as u32,
+                height: self.max_texture_size as u32,
                 depth_or_array_layers: 1,
             };
 
             context.queue.write_texture(
-                // Tells wgpu where to copy the pixel data
-                wgpu::ImageCopyTexture {
-                    texture: &self.texture,
+                wgpu::TexelCopyTextureInfo {
+                    texture: &self.color_texture,
                     mip_level: 0,
-                    origin: wgpu::Origin3d { x: 0, y: 0, z: 0 },
+                    origin: wgpu::Origin3d::ZERO,
                     aspect: wgpu::TextureAspect::All,
                 },
-                // The actual pixel data
-                &self.atlas.buffer,
-                // The layout of the texture
-                wgpu::ImageDataLayout {
+                &self.color_atlas.buffer,
+                wgpu::TexelCopyBufferLayout {
                     offset: 0,
-                    bytes_per_row: Some((self.max_texture_size * 4).into()),
-                    rows_per_image: Some((self.max_texture_size).into()),
+                    bytes_per_row: Some(
+                        self.max_texture_size as u32 * self.color_atlas.channels as u32,
+                    ),
+                    rows_per_image: Some(self.max_texture_size as u32),
                 },
                 texture_size,
             );
+
+            self.color_atlas.fresh = false;
+            self.color_atlas.dirty = false;
         }
-        self.atlas.fresh = false;
-        self.atlas.dirty = false;
     }
 }
 
-fn fill(
+struct FillParams {
     x: u16,
     y: u16,
     width: u16,
     _height: u16,
-    image: &[u8],
     target_width: u16,
-    target: &mut [u8],
-) -> Option<()> {
-    let channels = 4;
-    let image_pitch = width as usize * channels;
-    let buffer_pitch = target_width as usize * channels;
-    let mut offset = y as usize * buffer_pitch + x as usize * channels;
+    channels: usize,
+}
+
+fn fill(params: FillParams, image: &[u8], target: &mut [u8]) -> Option<()> {
+    let image_pitch = params.width as usize * params.channels;
+    let buffer_pitch = params.target_width as usize * params.channels;
+    let mut offset =
+        params.y as usize * buffer_pitch + params.x as usize * params.channels;
     for row in image.chunks(image_pitch) {
         let dest = target.get_mut(offset..offset + image_pitch)?;
         dest.copy_from_slice(row);
